/****************************************************************************\
  Plughole by M. A. "Slimo" Fincher. (mauls@cu.warwick.ac.uk)

  Block deletion algorithm by Harvey J. "Max" Thompson. (max@cs.warwick.ac.uk)
  Hi-score table subsytem by Harvey J "Max" Thompson (max@cs.warwick.ac.uk)

  Modified by Harvey Thompson, 30th July 2002, marked as PLUGHOLE_POSIX.
  Probably modified slightly by Harvey Thompson, sometime during 1992.
\****************************************************************************/

#include <stdlib.h>		/* for getenv() */
#ifdef PLUGHOLE_POSIX
#include <termios.h>
#include <unistd.h>
#else /* !PLUGHOLE_POSIX */
#include <sys/termios.h>
				/* Get yer vomit! */
#undef ECHO
#undef NL0
#undef NL1
#undef TAB0
#undef TAB1
#undef TAB2
#undef XTABS
#undef CR0
#undef CR1
#undef CR2
#undef CR3
#undef FF0
#undef FF1
#undef BS0
#undef BS1
#undef TOSTOP
#undef FLUSHO
#undef PENDIN
#undef NOFLSH
#endif /* !PLUGHOLE_POSIX */
#include <curses.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <math.h>
#include "global.h"
#include "score.h"

extern time_t timeofday;

int shipx, shipy;
int stage, alive, finished, score, dist, num_clear, rate;
int bullx[MAX_BULLS], bully[MAX_BULLS], bullf[MAX_BULLS];
int cornerx[4], cornery[4];

int vectorx[4] = { 1, 0, -1, 0 };
int vectory[4] = { 0, 1, 0, -1 };

int clearx[MAX_CLEARS], cleary[MAX_CLEARS], cleardrct[MAX_CLEARS];
int clearflag[MAX_CLEARS];

char bricks[NUM_BRICKS];
char *termtype;

#ifdef SYSV
#define usleep(us)	nap(us/1000)
#define srandom(x)	srand(x)
#define random()	rand()
#endif

/****************************************************************************\
 FUNCTION: checkoption.
 Receives a string, and checks to see if string has been passed as an argument
 to the program.
 Returns position if found, 0 if not.

\****************************************************************************/

int checkoption( numoptions, option, checkoption )

    char *checkoption, *option[];
    int numoptions;

{
    int i;

    for ( i=1 ; i<numoptions ; i++ )
	if ( ! strncmp ( option[i] , checkoption, strlen ( checkoption ) ) )
	    return ( i );
    return ( 0 );
}

/****************************************************************************\
 FUNCTION: safeopenscores.

\****************************************************************************/

int safeopenscores ()

{
    int fd;

    switch ( fd = openscore ( SCOREFILE, TIMEOUT, OPENSCORE ) ) {
    case -3:
	perror ( SCOREFILE );
	exit ( 1 );
	break;
    case -2:
	printf ( "Creating score file: %s\n", SCOREFILE );
	switch ( newscore ( SCOREFILE, TIMEOUT ) ) {
	case -2:
	    printf ( "Score file doesnt really exist?\n" );
    perror ( SCOREFILE );
	    exit ( 2 );
	    break;
	case -1:
	    printf ( "Score file does exist but is locked!\n" );
	    perror ( SCOREFILE );
	    exit ( 3 );
	    break;
	default:
	    break;
	}
	exit ( 0 );
	break;
    case -1:
	printf ( "Score file does exist but is locked!\n" );
	perror ( SCOREFILE );
	exit ( 4 );
	break;
    }
    return ( fd );
}

/****************************************************************************\
 FUNCTION: printscores.

\****************************************************************************/

void printscores ( option )

char *option;

{
    int num, dummy, fd, i;

    num = width ( option+2, DEFSCORES, & dummy );
    if ( num < 1 || num > STSIZE ) {
	printf ( "Phlegm off - thats a silly number!\n" );
	exit ( 1 );
    }
    fd = safeopenscores ();
    (void) printf ( ".-----.---------------------------.---------.---------.\n");
    (void) printf ( "| No. | Name                      | Score   | Level   |\n");
    (void) printf ( ">-----+---------------------------+---------+---------<\n");
    for ( i = 1; i <= num; i++ )
	(void) printf ( "| %3d %s\n",
		       i,
		       getscore ( fd, i, "| %-25n | %7s | %-7l |" ) );

    (void) printf ( "`-----^---------------------------^---------^---------'\n");
    (void) closescore ( fd );
}

/****************************************************************************\
 PROCEDURE: dohiscore.

\****************************************************************************/

void dohiscore ()

{
    int fd, rank;
    char *name;

    fd = safeopenscores ();
    if (   ( name = getenv ( "NAME" ) ) == NULL
	&& ( name = getenv ( "LOGNAME" ) ) == NULL
	&& ( name = getenv ( "USER" ) ) == NULL
	&& ( name = getlogin () ) == NULL)
	name = "A Complete Chicken";
    switch ( ( rank = addscore ( fd, name, score, stage ) ) ) {
    case -1:
	perror ( "Cheese up in score file! ");
	exit ( 1 );
	break;
    case 0:
	break;
    default:
	(void) move ( 10, 22 );
	(void) printw ( "You rank #%d in the hiscore table!\n", rank );
	(void) refresh ();
	(void) usleep ( 500000 );
	break;
    }
    (void) closescore ( fd );
}

/****************************************************************************\
 FLUSH THE FROGGING KEYBOARD YOU CHICKEN

\****************************************************************************/

void flush_keyboard ()

{
#ifdef PLUGHOLE_POSIX
    tcflush(0, TCIFLUSH);
#else /* !PLUGHOLE_POSIX */
    int out = 0;
    ioctl ( 0, TCFLSH, &out );
#endif /* !PLUGHOLE_POSIX */

#if 0
  fd_set fds;			/* Will contain only stdin */
  struct timeval timeout;	/* Will contain a zero timeout (poll) */
  static char buffer[LINELEN];	/* Will be used as the buffer to read(2) */
  int status;			/* Contains select(2)'s return value */

  while (1) {
    FD_ZERO (&fds);
    FD_SET (0, &fds);
    timeout.tv_sec = 0L;
    timeout.tv_usec = 0L;
    if (((status = select (1, &fds, (fd_set*) NULL, (fd_set*) NULL,
			   &timeout)) == -1) && (errno != EINTR))
      exit ( 1 );
    if ((status == 0) || (status == -1))
      break;			/* Can only be -1 if errno == EINTR */
    (void) read (0, buffer, LINELEN);
  }
#endif
}


/****************************************************************************\
 FUNCTION: pollkeyboard.
 Finds out whether there are any characters to be read in from stdin.
 It returns the following values:-
       0  : No char to be read in the past (sec+usec/10^6).
       1  : There are chars to be read.
      -1  : Phlegm up!

\****************************************************************************/

int pollkeyboard ( sec, usec )

int sec, usec;

{

    fd_set rd, wr, xp;
    struct timeval timeout;

    timeout.tv_sec = sec;
    timeout.tv_usec = usec;

    FD_ZERO ( &rd );
    FD_ZERO ( &wr );
    FD_ZERO ( &xp );

    FD_SET ( 0, &rd );

    switch ( select ( 32, &rd, &wr, &xp, &timeout ) ) {
    case -1:
	return ( -1 );
    case 0:
	return ( 0 );
    default:
	if ( FD_ISSET ( 0, &rd ) )
	    return ( 1 );
	else
	    return ( 0 );
    }
}

void remove_ship ()

{
    (void) move ( shipy, shipx );
    (void) addstr ( "    " );      /* no. spaces = SHIP_WIDTH */
    (void) refresh ();
}    

void initialise ()

{
    termtype = getenv ( "TERM" );

    stage      = 1;
    alive      = 1;
    finished   = 0;
    shipx      = 2;
    shipy      = 12;
    dist       = 0;
    rate       = INITIAL_RATE;
    num_clear  = 0;
}

void stage_change ()

{
    int t, i;
    float chick;

    num_clear = 0;
    for ( t=0 ; t<MAX_CLEARS ; t++ ) clearflag[t] = 0;
    for ( t=0 ; t<MAX_BULLS ; t++ )  bullf[t] = 0;
    
    for ( i=1 ; i<77 ; i+=2 ) {
	for ( t=1 ; t<22 ; t++ ) {
	    (void) move ( t, i );
	    (void) addstr ( "  @@" );
	}
	(void) refresh();
    }
    for ( t=0 ; t<SCREEN_DEPTH ; t++ ) {
	(void) move ( t, 76 );
	(void) addstr ( "  ||" );
    }

    remove_ship ();
    (void) refresh ();
    flush_keyboard();
    score += ( 100-shipx ) * 50;
    shipx = 2;
    shipy = 12;
    finished = 0;
    dist = 0;
    rate = INITIAL_RATE;

    chick = 100;
    for ( t=0 ; t<stage ; t++ ) chick *= STAGE_RATIO;

    rate = (int) (( chick * rate ) / 100);

    stage++;
    sleep ( 2 );
    return;
}

void put_score ( points, level )

    int points, level;
{
    extern char *form();
    (void) move ( 23, 6 );
    (void) addstr ( form ( "Score: %i", points ) );
    (void) move ( 23, 60 );
    (void) addstr ( form ( "Level: %i", level ) );
    (void) refresh ();
}

void kill_ship ()

{
    int count, t;
    float bx[16], by[16], vx[16], vy[16];
    char bits[18];
    int flag[16];

    (void) move ( shipy, shipx );
    (void) addstr ( "    " );
    (void) strcpy ( bits , ">C=-++--++--++--" );
    for ( t=0 ; t<16 ; t++ ) {
	flag[t] = 0;
	bx[t] = shipx;
 	by[t] = shipy;
 	vx[t] = ( (float) (random()%10000)/3700.0 + 0.06) ;
 	if ( (random()%10000) < 5000 ) vx[t] = -vx[t];
 	vy[t] = ( (float) (random()%10000)/10005.0 + 0.02);
 	if ( (random()%10000) < 5000 ) vy[t] = -vy[t];
    }
    while ( count < 16 ) {
	for ( t=0 ; t<16 ; t++ ) {
	    if ( flag[t] == 0 ) {
		(void) move ( (int) (by[t]), (int) (bx[t]) );
		(void) addch ( ' ' );

 		bx[t] = bx[t] + vx[t];
 		if ( ( bx[t] < 1 ) || ( bx[t] > 75 ) ) {
		    flag[t] = 1;
		    count++;
		}
		by[t] = by[t] + vy[t];
		vy[t] += 0.04;
		if ( ( by[t] < 3 ) || ( by[t] > 21 ) ) {
		    flag[t] = 1;
		    count++;
		}
		(void) move ( (int) (by[t]), (int) (bx[t]) );
		(void) addch ( bits[t] );
	    }
	}
	(void) refresh ();
	(void) usleep ( 50000 );
    }
}

void place_blocks ()

{
    int t, i, j, g, k, number, choice, y, flag;
    char thing;
    
    number = 40;

    for ( t=0 ; t<number ; t++ ) {
	choice = random() % NUM_BLOCKS;
	y = random() % 16 + 1;
	i = 0;
	flag = 1;
	while ( ( i<LAST_CLMN ) && ( flag == 1 ) ) {
	    i++;
	    for ( j=0 ; j<BLOCK_DEPTH ; j++ ) {
		(void) move ( y+j, i );
		thing = inch();
		if ( ( thing != ' ' ) && ( blocks[choice][j][BLOCK_WIDTH-1] != ' ' ) ) { 
		    flag = 0;
		    break;
		}
	    }
	}
	for ( j=0 ; j<BLOCK_DEPTH ; j++ ) {
	    for ( k=0 ; k<BLOCK_WIDTH ; k++ ) {
		(void) move ( y+j, i-BLOCK_WIDTH+k );
		if ( blocks[choice][j][k] == '#' ) {
		    if ( t == 0 ) (void) addch ( ':' );
		    else (void) addch ( bricks[( t % NUM_BRICKS )] );
		}
	    }
	}
	(void) refresh();
    }
}

int clear_blocks () 

{
    int t;
    char thing;

    if ( num_clear == 0 ) return 0;
    for ( t=0 ; t<MAX_CLEARS ; t++ ) {
	if ( clearflag[t] > 0 ) {
	
	    (void) move ( cleary[t], clearx[t] );
	    (void) addch ( ' ' );
	    score += 15;
	    
	    (void) move ( cleary[t]+vectory[cleardrct[t]], 
			 clearx[t]+vectorx[cleardrct[t]] );
	    thing = inch ();
	    if ( thing == '*' ) {
		cleary[t] += vectory[cleardrct[t]];
		clearx[t] += vectorx[cleardrct[t]];
		clearflag[t] = 1;
	    } else {
		cleardrct[t] = ( cleardrct[t] + 1 ) % 4;
		clearflag[t]++;
		if ( clearflag[t] == 6 )  {
		    clearflag[t] = 0;
		    num_clear--;
		}
	    }
	}
    }
    (void) refresh ();
    return 1;
}

void place_ship ()

{
    (void) move ( shipy, shipx );
    (void) addstr ( ship );
    (void) refresh ();
}


void end_game ()
{
    finished = 1;
    alive = 0;
    kill_ship ();
}

void check_ship ()

{
    int t;
    
    for ( t=0 ; t<SHIP_WIDTH ; t++ ) {
	(void) move ( shipy, shipx+t );
	if ( inch() != ' ' ) {
	    end_game ();
	    return;
	}
    }
    place_ship();
}

void check_new_block ( v, u )
    int v, u;
{
    int t;
    int ml, mr, md;
    int i, j;
    int f = 0;
    char last = ' ', curr;

    if ( find_block ( v, 79-u, &ml, &mr, &md ) ) {

	for ( i = v - ml; i <= v + mr; i++ )
	    for ( j = u ; j <= u + md; j++  ) {
		move ( i, j );
		if ( (curr = inch()) == ':' ) 
		    finished = 1;
		if ( last == ' ' )
		    last = curr;
		if ( last != curr )
		    f = 1;
		addch ( '*' );
	    }
	rate--;
	if ( f )
	    score += (ml+mr+1)*(md+1)*5;

	for ( t=0 ; t<MAX_CLEARS ; t++ ) {
	    if ( clearflag[t] == 0 ) {
		clearx[t] = u;
		cleary[t] = v - ml;
		cleardrct[t] = 3;
		clearflag[t] = 1;
		num_clear++;
		break;
	    }
	}
    }
}

void move_bullets ()

{
    int t;
    char thing;
    
    for ( t=0 ; t<MAX_BULLS ; t++ ) {
	if ( bullf[t] == 1 ) {
	    (void) move ( bully[t], bullx[t] );
	    (void) addch ( ' ' );
	    bullx[t]++;
	    (void) move ( bully[t], bullx[t] );
	    thing = inch();
	    if ( ( thing != ' ' ) && ( thing != 'o' ) ) {
		if ( thing == '*' ) {
		    bullf[t] = 0;
		    refresh ();
		} else {
		    (void) move ( bully[t], bullx[t]-1 );
		    (void) addch ( thing );
		    score += 5;
		    check_new_block ( bully[t], bullx[t]-1 );
		    bullf[t] = 0;
		}
	    } else {
		(void) move ( bully[t], bullx[t] );
		(void) addch ( 'o' );
	    }
	    (void) refresh ();
	}
    }
}

void advance_ship ()

{
#ifndef DEBUG
    dist++;
    if ( dist > rate ) {
	remove_ship ();
	shipx++;
	check_ship ();
	dist = 0;
    }
#endif
}

void move_ship ()

{
    int keyflag, mul, t;
    char key;

    switch ( pollkeyboard ( 0, 6000 ) ) {
    case -1:
	(void) addstr ( "Phlegm up, Phlegm up! ERROR 3.chicken" );
	exit ( 1 );
	break;
	
    case 0:
	(void) usleep ( 7500 );
	keyflag = 0;
	break;

    default:
	keyflag = 1;
	if ( read ( 0, &key, 1 ) != 1 ) {
	    (void) addstr ( "Phlegm up, Phlegm up, ERROR chicken.2" );
	    exit ( 1 );
	}
	flush_keyboard();
    }
    
    if ( keyflag ) {
	if ( ( key == 'q' ) && ( shipy > 1 ) ) {
	    remove_ship ();
	    shipy--;
	    check_ship ();
	    return;
	}
	if ( ( key == 'a' ) && ( shipy < 21 ) ) {
	    remove_ship ();
	    shipy++;
	    check_ship ();
	    return;
	}
	if ( ( key == 'S' ) )
	    end_game ();

	if ( key == ' ' ) {
	    for ( t=0 ; t<MAX_BULLS ; t++ ) {
		if ( bullf[t] == 0 ) {
		    bullf[t] = 1;
		    bully[t] = shipy;
		    bullx[t] = shipx + SHIP_WIDTH;
		    break;
		}
	    }
	}
	if ( key == 'l' ) (void) wrefresh ( curscr );	    
    }
}

void set_up_screen ()

{
    time_t *big;
    int t;

    (void) srandom ( time(&big) );
    (void) initscr();
    (void) cbreak();
    (void) noecho();
    (void) move ( 0, 0 );
    (void) addstr ( "<<<<><<>><<<>>><<<<>>>><<<<<>>>>><<<<<<>>>>>><<<<<>>>>><<<<>>>><<<>>><<>><>>>" );
    (void) move ( 22, 0 );
    (void) addstr ( "<<<<><<>><<<>>><<<<>>>><<<<<>>>>><<<<<<>>>>>><<<<<>>>>><<<<>>>><<<>>><<>><>>>" );
    (void) move ( 23, 31 );
    (void) addstr ( "P*L*U*G  H*O*L*E" );
    if ( strncmp ( termtype, "tvi", 3 ) == 0 )
         (void) addstr ( " " ); 
    for ( t=0 ; t<SCREEN_DEPTH ; t++ ) {
	(void) move ( t, 78 );
	(void) addch ( '|' );
    }
    put_score ( 0, 1 );
    (void) refresh();
}
    
main ( argc, argv )

    int argc; char *argv[];

{

    int t, scoreopt;


    for ( t=0 ; t<MAX_BULLS ; t++ ) bullf[t] = 0;
    for ( t=0 ; t<MAX_CLEARS ; t++ ) clearflag[t] = 0;

    (void) strcpy ( bricks, "0ABCDE#FGHIJ&KLMNO%PQRST$UVWXYZ@" );

    if ( scoreopt = checkoption( argc, argv, "-s" ) ) {
	printscores ( argv[scoreopt] );
	exit ( 1 );
    }

    initialise ();
    set_up_screen ();
    while ( alive == 1 ) {
	place_blocks ();
	place_ship ();
	while ( ! finished ) {
	    move_ship ();
	    move_bullets ();
	    if ( clear_blocks () == 0 )
		advance_ship ();
	    put_score ( score, stage );
	}
	if ( alive == 1 ) stage_change();
    }
    dohiscore ();  
    if ( strncmp ( termtype, "tvi", 3 ) == 0 )
	(void) addstr ( " " ); 
    (void) refresh();
    endwin ();
}

